# #
# $Date: 1996/06/19 15:28:06 $  $Author: frankp $   $Revision: 1.10 $ #
# #
# frankp, 14.06.1995 #

#++
  eigenvec.mu

	linalg::eigenVectors(A)  --  computes the eigenvectors of matrix A

	A: square matrix

	eigenVectors(A) computes the eigenvalues and
	eigenvectors of the matrix A.

	First all the eigenvalues of the matrix are computed using the 
	function linalg::eigenValues. 

	After this, for any eigenvalue u the matrix equation
	(I*u - A)*X = 0, where I denotes the identity matrix, is solved 
	using the Gauss-Jordan algorithm (linalg::gaussJordan).
	If A is not of category IntegralDomain, then an error message 
	will be given.

	The result is a list of sublists. Each sublist contains
	in that order an eigenvalue u, its algebraic multiplicity
	and a list of eigenvectors corresponding to u.

	When the sum of the algebraic multiplicities are not equal to
        the dimension of A then FAIL is returned.
++#

linalg::eigenVectors := proc(A)
    name linalg::eigenVectors;
    option remember;
    local n, s, i, k, B, r, Aplus, ev, En;
begin
    if testargs() then
	if args(0) <> 1 then
	    error("wrong no of args")
	end_if;
	if A::hasProp( Cat::MatrixCat ) <> TRUE then
	    error("expecting a square matrix")
	end_if;
	if not (A::coeffRing)::hasProp( Cat::Field ) then
	    error("expecting matrix over a field")
	end_if;
	n := A::dimen(A);
	if op(n,1) <> op(n,2) then
	    error("expecting a square matrix")
	end_if
    end_if;

    userinfo(1,"compute system of eigenvectors");

    n := op(A::dimen(A),1);
    userinfo(2,"compute the eigenvalues");
    s := linalg::eigenValues(A,hold(Multiple));

    if _plus(op(s,[i,2]) $ i=1..nops(s)) <> n then
	userinfo(2,"algebraic multiplicity <> geometric multiplicity");
	return( FAIL )
    end_if;

    A:= A::negate(A);
    Aplus:= A::_plus;
    En:= A::newThis(n,n,subs(fun(hold(XX)),hold(XX)=(A::coeffRing)::one),hold(Diagonal));
    ev:= [];
    for k from 1 to nops(s) do
	B:= En;
	(B[i,i]:= op(s,[k,1])) $ i=1..n;
	userinfo(2,"compute kernel of characteristic matrix");
 	r:= linalg::nullSpace( Aplus(B,A) );	
	ev:= append(ev,[op(s,[k,1]),op(s,[k,2]),r])
    end_for;

    ev
end_proc:

# end of file #
