/******************************************************************************/
/* DATEI:       MMT_tool.c                                                    */
/* AUTOR:       Andreas Sorgatz                                               */
/* ERSTELLT:    02/03/93                                                      */
/* GEAENDERT:   01/12/94                                                      */
/* INHALT:      Definitionen zur einfacheren Modulprogrammierung.             */
/*                                                                            */
/* Die Funktionen sortiert nach dem Kontext ihrer Verwendung:                 */
/*                                                                            */
/* MMT_check_param      - Prueft Typ des n-ten F-Arguments --> ggf. F-Abbruch */
/* MMT_check_param_domext  - Prueft n-tes F.Arg auf DOM_EXT des Moduldomains  */
/* MMT_check_params     - Prueft Anzahl  der   F-Argumente --> ggf. F-Abbruch */
/* MMT_error            - Beendet die Funktion mit Fehlermeldung und -abbruch */
/* MMT_init             - Muss jede Modulfunktion einleiten                   */
/* MMT_no_write_ass_mem - Verhindert das automatische Eintragen ins AssMem    */
/* MMT_param            - Liefert das n-te F-Argument                         */
/* MMT_params           - Liefert Anzahl der F-Argumente                      */
/* MMT_return           - Beendet die Funktion mit einem Rueckgabewert        */
/*                                                                            */
/* MMT_break            - Wie MMT_error, ABER OHNE REINITIALISIERUNGEN        */
/* MMT_puts             - Ausgabe: MuPAD-Ausgabe wie puts                     */
/*                                                                            */
/* MMT_cat              - Liefert die Categorie eines Objektes                */
/* MMT_copy             - Liefert eine neue Instanz des gegebenen Objektes    */
/* MMT_count            - Liefert die Anzahl direkter Unterobj. eines Obj.    */
/* MMT_free             - Loescht das gegebene Objekt (bzw. eine Instanz)     */
/* MMT_gop              - Liefert den n-ten S_Pointer eines S_Pointers        */
/* MMT_pooldelete       - Entsichert einen Zeiger in einen S_Pointer          */
/* MMT_poolinsert       - Sichert    einen Zeiger in einen S_Pointer          */
/* MMT_sig              - Berechnet und setzt die Signatur eines Objektes     */
/* MMT_sop              - Besetzt den n-ten S_Pointer eines S_Pointers        */
/* MMT_spec             - Liefert den special-Eintrag  eines Objektes         */
/*                                                                            */
/* MMT_c2m_bool         - Wandelt: C-       in MuPAD-bool       (**)          */
/* MMT_c2m_double       - Wandelt: C-       in MuPAD-double     (**)          */
/* MMT_c2m_ident        - Wandelt: C-String in MuPAD-Identifier (**)          */
/* MMT_c2m_long         - Wandelt: C-       in MuPAD-long                     */
/* MMT_c2m_string       - Wandelt: C-       in MuPAD-string     (**)          */
/* MMT_m2c_bool         - Wandelt: MuPAD-   in C    -bool                     */
/* MMT_m2c_double       - Wandelt: MuPAD-   in C    -double                   */
/* MMT_m2c_ident        - Wandelt: MuPAD-Identifier in C-string (**)          */
/* MMT_m2c_ident_adr    - Liefert: Adresse des C-string eines MuPAD-Ident.    */
/* MMT_m2c_long         - Wandelt: MuPAD-   in C    -long                     */
/* MMT_m2c_string       - Wandelt: MuPAD-   in C    -string     (**)          */
/* MMT_m2c_string_adr   - Liefert: Adresse des C-string eines MuPAD-strings   */
/*                                                                            */
/* MMT_cat2pari         - Wandelt: MuPAD-Categorie(Zahl) in PARI-Objekt(GEN)  */
/* MMT_pari2cat         - Wandelt: PARI-Objekt(GEN) in MuPAD-Kategorie(Zahl)  */
/*                                                                            */
/* MMT_dom_entry        - Liest einen Eintrag des Moduldomains aus            */
/* MMT_exec_dom         - Liefert das Moduldomain zum gegebenen EXEC-Knoten   */
/* MMT_this_dom         - Liefert das Moduldomain                             */
/* MMT_is_ext_of_dom    - Prueft Zugehoerigkeit eines DOM_EXT zu Domain       */
/*                                                                            */
/*                                                                            */
/* MMT_new_cat_expr     - Erzeugt ein DOM_EXPR                                */
/* MMT_new_cat_expr2    - Erzeugt ein DOM_EXPR mit Eintraegen                 */
/* MMT_new_cat_ext      - Erzeugt ein DOM_EXT                                 */
/* MMT_new_cat_list     - Erzeugt ein DOM_LIST                                */
/*                                                                            */
/* MMT_call_eval_func   - Aufruf einer _eval_ Funktion                        */
/*                                                                            */
/* MMT_read_ident       - Liefert den Wert eines Bezeichners                  */
/* MMT_read_ident_c     - Liefert den Wert eines Bezeichners                  */
/* MMT_write_ident      - Zuweisung an einen Bezeichner                       */
/* MMT_write_ident_c    - Zuweisung an einen Bezeichner                       */
/*                                                                            */
/* MMT_equal            - Vergleich:  a == b   (beliebige Objekte)            */
/*                                                                            */
/* MMT_cmp_ident_c      - Vergleich:  strcmp(MuPADstring,Cstring)             */
/* MMT_cmp_string_c     - Vergleich:  strcmp(MuPADstring,Cstring)             */
/*                                                                            */
/* MMT_eq               - Vergleich:  a == b                                  */
/* MMT_ne               - Vergleich:  a != b                                  */
/* MMT_ge               - Vergleich:  a >= b                                  */
/* MMT_gt               - Vergleich:  a >   b                                 */
/* MMT_is_minus_one     - Prueft, ob das Objekt die Zahl -1 ist               */
/* MMT_is_number        - Prueft, ob das Objekt eine Zahl ist                 */
/* MMT_is_one           - Prueft, ob das Objekt die Zahl  1 ist               */
/* MMT_is_zero          - Prueft, ob das Objekt die Zahl  1 ist               */
/* MMT_le               - Vergleich:  a <  b                                  */
/* MMT_lt               - Vergleich:  a <= b                                  */
/*                                                                            */
/* MMT_abs              - Berechnet: Absolutwert einer komplexen Zahl         */
/* MMT_add              - Berechnet: Addition                                 */
/* MMT_add_long         - Berechnet: Addition einer Zahl vom C-Typ long       */
/* MMT_dec              - Berechnet: Dekrementierung                          */
/* MMT_div              - Berechnet: Division                                 */
/* MMT_div_int          - Berechnet: ganzzahlige Division                     */
/* MMT_gcd              - Berechnet: groessten gemeinsamen Teiler             */
/* MMT_inc              - Berechnet: Inkrementierung                          */
/* MMT_lcm              - Berechnet: kleinstes gemeinsames Vielfaches         */
/* MMT_mod              - Berechnet: Rest der ganzzahligen Division           */
/* MMT_mult             - Berechnet: Multiplikation                           */
/* MMT_power            - Berechnet: Potentzierung                            */
/* MMT_sub              - Berechnet: Subtraktion                              */
/*                                                                            */
/* MMT_complex_i        - Zahl:   i      (*)                                  */
/* MMT_fail             - Wert:   FAILED (*)                                  */
/* MMT_false            - Bool:   FALSE  (*)                                  */
/* MMT_half             - Zahl:   1/2    (*)                                  */
/* MMT_minus_one        - Zahl:   -1     (*)                                  */
/* MMT_minus_two        - Zahl:   -2     (*)                                  */
/* MMT_nil              - Wert:   NIL    (*)                                  */
/* MMT_null             - Wert:   NULL   (*)                                  */
/* MMT_one              - Zahl:   1      (*)                                  */
/* MMT_true             - Bool:   TRUE   (*)                                  */
/* MMT_two              - Zahl:   2      (*)                                  */
/* MMT_zero             - Zahl:   0      (*)                                  */
/*                                                                            */
/*                                                                            */
/* (*)   Muss zum Gebrauch ggf. kopiert (und spaeter ggf. geloescht) werden.  */
/*       Dies geschieht mit der Funktion `MMT_copy' (vgl. auch `MMT_free').   */
/* (**)  Erstellt eine Kopie,  die Vorlage kann geloescht werden.  Die Kopie  */
/*       muss vom Benutzer am Fkt.-Ende ggf. geloescht werden.                */
/*                                                                            */
/******************************************************************************/

#ifndef __MMT_tool__
#define __MMT_tool__

/*** Konstanten zur Modul-Programmierung **************************************/

#define MMT_STORAGE	(MMMCglobal)	   /*** Speichertyp aller Variablen ***/

#define MMT_NOP		0L		              /*** Option: -        ***/
#define MMT_HOLD	1L		              /*** Option: hold     ***/
#define MMT_REMEMBER	2L                            /*** Option: remember ***/

#define MMT_ANY		MEVC_NO_TYPE_CHECK        /*** Typen fuer Fkt.-Wert ***/

#define MMT_EVAL	MGSC_FLAT_ARG                         /* Standardwert */

#define MMT_NO_SUPPORT  "Sorry, not supported on this system"

#define MMT_NOT_UNIX    ( defined MACINTOSH || defined MSDOS )


/*** Konstanten zur internen Steuerung ****************************************/

#define MMT_FREE_S	1L                              /*** fuer MMTV_ctrl ***/


/*** Globale Variablen fuer das Modul *****************************************/

static char		 MMTV_buf[4096];
static char		*MMTV_name = "unknown";
static long		 MMTV_l, MMTV_ctrl = 0L;
static double		 MMTV_d;
static S_Pointer	 MMTV_r, MMTV_s, MMTV_t, MMTV_u, MMTV_v, MMTV_w;
static S_Pointer	 MMTV_s_orig = MMMNULL;
			

/*** Prototypen ***************************************************************/

static void	 MMT_change	    ( S_Pointer *p );
static S_Pointer MMT_copy	    ( S_Pointer  p );
static S_Pointer MMT_free           ( S_Pointer  p );
static void	 MMT_inactiv	    ( S_Pointer *p );   

static S_Pointer MMT_new_cat_expr2  ( long       num Mva_arglist );
static S_Pointer MMT_new_cat_ext    ( long count, long size );
static S_Pointer MMT_call_eval_func ( char	*name, long num Mva_arglist );
static S_Pointer MMT_call_mf        ( char      *mname,
                                      char      *fname,
                                      S_Pointer  s,
                                      long       prev_func,
                                      long       eval_type,
                                      S_Pointer  exec
                                    );
static char	*MMT_m2c_string     ( S_Pointer  mstr );
static char	*MMT_m2c_string_adr ( S_Pointer  mstr );

static void	 MT_printf	    ( char      *format Mva_arglist );

static S_Pointer MMT_read_ident	    ( S_Pointer ident );
static S_Pointer MMT_read_ident_c   ( char      *cstr );
static void	 MMT_write_ident    ( S_Pointer ident, S_Pointer value );
static void	 MMT_write_ident_c  ( char      *cstr, S_Pointer value );

static long      MMT_is_ext_of_dom  ( S_Pointer domain, S_Pointer value  );
static S_Pointer MMT_dom_entry      ( S_Pointer domain, S_Pointer string );
static S_Pointer MMT_dom_entry_c    ( S_Pointer domain, char *string     );


/*** MAKROS *******************************************************************/

/*** Definments fuer die Namen der originalen Modulfunktionsargumente. Nur ****/
/*** diese Namen (Makros) duerfen im folgenden verwendet werden.           ****/

#define MMTV__s			s
#define MMTV__prev_func		prev_func
#define MMTV__eval_type		eval_type
#define MMTV__exec		exec



/******************************************************************************/
/* NAME:        MMT_abs                                                       */
/* PARAMETER:   p = (S_Pointer) Ist S_Pointer vom Typ CAT_COMPLEX             */
/* FUNKTION:    Das Makro liefert das Ergebnis "|p|".                         */
/******************************************************************************/
#define MMT_abs( p )							\
(									\
	MCA_abs(p)							\
)

/******************************************************************************/
/* NAME:        MMT_add                                                       */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 + p2".                     */
/******************************************************************************/
#define MMT_add( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_PLUS)					\
)

/******************************************************************************/
/* NAME:        MMT_add_long                                                  */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer aus der Typenmenge              */
/*                  { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }     */
/*              l = (long) Eine Zahl vom C-Typ long                           */
/* FUNKTION:    Das Makro liefert das Ergebnis "p + l".                       */
/******************************************************************************/
#define MMT_add_long( p, l )						\
(									\
	MCA_plus_minus((p),(l))						\
)

/******************************************************************************/
/* NAME:        MMT_break                                                     */
/* PARAMETER:   text = Fehlermeldung                                          */
/* FUNKTION:    Das Makro beendet die Modulfunktion  (OHNE Reinitialisierung) */
/*              mit der gegebenen Fehlermeldung und einem Fehlerabbruch.      */
/******************************************************************************/
#define MMT_break( text )						\
{									\
	sprintf( MMTV_buf, "%s [%s]", text, MMTV_name );		\
	MEV_set_error( MEVC_USER_DEFINED_ERROR, MMTV_buf );		\
}

/******************************************************************************/
/* NAME:        MMT_cat                                                       */
/* PARAMETER:   p = (S_Pointer) Ein MuPAD-Datum                               */
/* FUNKTION:    Liefert die Categorie des S_Pointers p.                       */
/******************************************************************************/
#define MMT_cat(p)	 						\
(									\
	MMTV_t = (p),							\
	MMTV_l = MMMZ(&MMTV_t,CATEGORY,MMMread)->cat,			\
	MMT_inactiv(&MMTV_t),						\
	MMTV_l								\
)

/******************************************************************************/
/* NAME:        MMT_cat2pari                                                  */
/* PARAMETER:   p = (S_Pointer) Eine MuPAD-Zahl                               */
/* FUNKTION:    Wandelt die Categorie (Zahl) in ein PARI-PObjekt (GEN).       */
/******************************************************************************/
#define MMT_cat2pari(p)							\
(									\
	MCA_cat_to_pari((p))						\
)

/******************************************************************************/
/* NAME:        MMT_check_param                                               */
/* PARAMETER:   n   = (long) Paramaterindex aus [0 .. MMTD_MMTD_NOPS()]       */
/*              cat = eine Categorie                                          */
/* FUNKTION:    Das Makro prueft, ob der n-te Funktionsparameter vom Typ der  */
/*              gegebenen Categorie ist und bricht die Funktion mit einer Fe- */
/*              hlermeldung ab, wenn dies nicht der Fall ist.                 */
/******************************************************************************/
#define MMT_check_param( n, cat )					\
{									\
	if( MMT_cat(MMT_param(n)) != cat ) {				\
		MMT_exit();						\
		MEV_set_error( MEVC_ILLEGAL_ARGUMENT, MMTV_name );	\
	}								\
}

/******************************************************************************/
/* NAME:        MMT_check_param_domext                                        */
/* PARAMETER:   n   = (long) Paramaterindex aus [0 .. MMTD_MMTD_NOPS()]       */
/* FUNKTION:    Das Makro prueft, ob der n-te Funktionsparameter ein CAT_EXT  */
/*              des aktuellen Domains ist. Die Funktion bricht mit einer Fe-  */
/*              hlermeldung ab, wenn dies nicht der Fall ist.                 */
/******************************************************************************/
#define MMT_check_param_domext( n )					\
{									\
	if( !MMT_is_ext_of_dom(MMT_this_dom(),MMT_param(n)) ) {		\
		MMT_exit();						\
		MEV_set_error( MEVC_ILLEGAL_ARGUMENT, MMTV_name );	\
	}								\
}

/******************************************************************************/
/* NAME:        MMT_check_params                                              */
/* PARAMETER:   n = (long) Parameteranzahl                                    */
/* FUNKTION:    Das Makro prueft, ob die Funktion genau n Parameter hat. Ist  */
/*              dies nicht der Fall,  so wird die Funktion mit einer Fehler-  */
/*              meldung abgebrochen.                                          */
/******************************************************************************/
#define MMT_check_params( n )	 					\
{									\
	if( MMT_params() != (long) (n) ) {				\
		MMT_exit();						\
		MEV_set_error( MEVC_WRONG_NUMBER_OF_ARGS, MMTV_name );	\
	}								\
}

/******************************************************************************/
/* NAME:        MMT_cmp_ident_c                                               */
/* PARAMETER:   p = (S_Pointer) Ein MuPAD-Bezeichner                          */
/*              s = (char*)     Ein C-String                                  */
/* FUNKTION:    Liefert das Ergebnis des Stringvergleiches mit 'strcmp'.      */
/******************************************************************************/
#define MMT_cmp_ident_c( p, s )						\
(									\
	strcmp( MMT_m2c_ident_adr(p), (s) )				\
)

/******************************************************************************/
/* NAME:        MMT_cmp_string_c                                              */
/* PARAMETER:   p = (S_Pointer) Ein MuPAD-String                              */
/*              s = (char*)     Ein C-String                                  */
/* FUNKTION:    Liefert das Ergebnis des Stringvergleiches mit 'strcmp'.      */
/******************************************************************************/
#define MMT_cmp_string_c( p, s )					\
(									\
	strcmp( MMT_m2c_string_adr(p), (s) )				\
)

/******************************************************************************/
/* NAME:        MMT_complex_i                                                 */
/* FUNKTION:    Die komplexe Zahl `i'.                                        */
/******************************************************************************/
#define MMT_complex_i							\
	MCAC_ci

/******************************************************************************/
/* NAME:        MMT_count                                                     */
/* PARAMETER:   p = (S_Pointer) Ein MuPAD-Datum                               */
/* FUNKTION:    Liefert die Anzahl der S_Pointer-Eintraege des S_Pointers p.  */
/******************************************************************************/
#define MMT_count(p)	 						\
(									\
	MMMcounter(p)							\
)

/******************************************************************************/
/* NAME:        MMT_c2m_bool                                                  */
/* PARAMETER:   i = (int) Ein boolscher Wert in C: 0->FALSE, sonst->TRUE      */
/* FUNKTION:    Liefert den boolschen Wert i als MuPAD-Objekt (CAT_BOOL).     */
/******************************************************************************/
#define MMT_c2m_bool( i )						\
(									\
	MMTV_l = (long)(i),						\
	MMT_copy( (MMTV_l ? MMT_true : MMT_false) )			\
)

/******************************************************************************/
/* NAME:        MMT_c2m_double                                                */
/* PARAMETER:   cdouble = (double) Ein C-double                               */
/* FUNKTION:    Liefert einen MuPAD-float (CAT_FLOAT) zum C-double.           */
/******************************************************************************/
#define MMT_c2m_double( cdouble )					\
(									\
	MMTV_d = (double)cdouble,					\
	MCA_new_cat_pari(MMT_STORAGE,MCA_double_to_pari(cdouble))	\
)

/******************************************************************************/
/* NAME:        MMT_c2m_ident                                                 */
/* PARAMETER:   cstr = Ein C-String                                           */
/* FUNKTION:    Liefert einen MuPAD-Identifier (CAT_IDENT) zum C-String       */
/******************************************************************************/
#define MMT_c2m_ident(cstr)						\
(									\
	MTR_new_cat_ident(MMT_STORAGE,cstr)				\
)

/******************************************************************************/
/* NAME:        MMT_c2m_long                                                  */
/* PARAMETER:   clong = Ein C-long                                            */
/* FUNKTION:    Liefert einen MuPAD-Integer (CAT_INT) zum C-long              */
/******************************************************************************/
#define MMT_c2m_long( clong )						\
(									\
	MMTV_l = (long)(clong),						\
	MCA_set_long(MMTV_l)						\
)

/******************************************************************************/
/* NAME:        MMT_c2m_string                                                */
/* PARAMETER:   cstr = Ein C-String                                           */
/* FUNKTION:    Liefert einen MuPAD-String (CAT_STRING) zum C-String          */
/******************************************************************************/
#define MMT_c2m_string(cstr)						\
(									\
	MTR_new_cat_string(MMT_STORAGE,cstr)				\
)

/******************************************************************************/
/* NAME:        MMT_dec                                                       */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer aus der Typenmenge              */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "--p".                         */
/******************************************************************************/
#define MMT_dec( p )							\
(									\
	MCA_mm(p)							\
)

/******************************************************************************/
/* NAME:        MMT_div                                                       */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 / p2".                     */
/******************************************************************************/
#define MMT_div( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_DIV)					\
)

/******************************************************************************/
/* NAME:        MMT_div_int                                                   */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "[p1 / p2]".                   */
/******************************************************************************/
#define MMT_div_int( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_DIVG)					\
)

/******************************************************************************/
/* NAME:        MMT_eq                                                        */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 == p2".                    */
/******************************************************************************/
#define MMT_eq( p1, p2 )						\
(									\
	MCA_eq(p1,p2)							\
)

/******************************************************************************/
/* NAME:        MMT_equal                                                     */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind beliebige S_Pointer                 */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 == p2".                    */
/******************************************************************************/
#define MMT_equal( p1, p2 )						\
(									\
	MCM_equal(p1,p2)						\
)

/******************************************************************************/
/* NAME:        MMT_error                                                     */
/* PARAMETER:   text = Fehlermeldung                                          */
/* FUNKTION:    Das Makro beendet die Modulfunktion  (nach Reinitialisierung) */
/*              mit der gegebenen Fehlermeldung und einem Fehlerabbruch.      */
/******************************************************************************/
#define MMT_error( text )						\
{									\
	MMT_exit();							\
	sprintf( MMTV_buf, "%s [%s]", text, MMTV_name );		\
	MEV_set_error( MEVC_USER_DEFINED_ERROR, MMTV_buf );		\
}
	
/******************************************************************************/
/* NAME:        MMT_exec_dom                                                  */
/* PARAMETER:   exec = (DOM_EXEC) EXEC-Knoten der Modulfunktion               */
/* FUNKTION:    Die Funktion liest den Verweis auf das Moduldomain aus dem    */
/*              EXEC-Eintrag hinter der remember-table (keine Kopie!!).       */
/******************************************************************************/
#define MMT_exec_dom( exec )						\
(									\
	MMTV_s = (exec),						\
	MMTV_t = MDM_modul_domain(MMTV_s),				\
	MMT_inactiv(&MMTV_s),						\
	MMTV_t								\
)

/******************************************************************************/
/* NAME:        MMT_exit                                                      */
/* PARAMETER:   -                                                             */
/* FUNKTION:    Das Makro loescht eine ggf. durch  `MMT_init()'  erstellte    */
/*              Kopie der Funktionsparameter und setzt Flags zurueck.         */
/******************************************************************************/
#define MMT_exit()							\
{									\
	MMMinactiv(&MMTV__s);						\
	MMMfree(MMTV__s);						\
	MMTV_ctrl   = 0;						\
	MMTV_s_orig = MMMNULL;						\
}

/******************************************************************************/
/* NAME:        MMT_fail                                                      */
/* FUNKTION:    Der Wert FAILED.                                              */
/******************************************************************************/
#define MMT_fail							\
	MUPV_CAT_FAILED

/******************************************************************************/
/* NAME:        MMT_false                                                     */
/* FUNKTION:    Der boolsche Wert `FALSE'.                                    */
/******************************************************************************/
#define MMT_false							\
	MUPV_CAT_BOOL_FALSE

/******************************************************************************/
/* NAME:        MMT_gcd                                                       */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "ggt(p1,p2)".                  */
/******************************************************************************/
#define MMT_gcd( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_GCD)					\
)

/******************************************************************************/
/* NAME:        MMT_ge                                                        */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 >= p2".                    */
/******************************************************************************/
#define MMT_ge( p1, p2 )						\
(									\
	(int) MCA_ge(p1,p2)						\
)

/******************************************************************************/
/* NAME:        MMT_gop                                                       */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer mit Unterobjekten               */
/*              n = (long) Index aus [0 .. MMT_count(s)]                      */
/* FUNKTION:    Liefert den n-ten S_Pointer aus dem S_Pointer p.              */
/******************************************************************************/
#define MMT_gop( p, n )	 						\
(									\
	MMTV_u = (p),							\
	MMTV_l = (long)(n),						\
	MMTV_v = *MMMP(&MMTV_u,MMTV_l,MMMread),				\
	MMT_inactiv(&MMTV_u),						\
	MMTV_v								\
)

/******************************************************************************/
/* NAME:        MMT_pooldelete                                                */
/* PARAMETER:   p = Adresse eines Zeiger in einen S_Pointer                   */
/* FUNKTION:    Entsichert diesen Zeiger gegen das Verschieben des Speichers. */
/******************************************************************************/
#define MMT_pooldelete( p )	 					\
{									\
	MMMpooldelete(p)						\
}

/******************************************************************************/
/* NAME:        MMT_poolinsert                                                */
/* PARAMETER:   p = Adresse eines Zeiger in einen S_Pointer                   */
/* FUNKTION:    Sichert diesen Zeiger gegen das Verschieben des Speichers.    */
/******************************************************************************/
#define MMT_poolinsert( p )	 					\
{									\
	MMMpoolinsert(p)						\
}

/******************************************************************************/
/* NAME:        MMT_gt                                                        */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 > p2".                     */
/******************************************************************************/
#define MMT_gt( p1, p2 )						\
(									\
	(int) MCA_gt(p1,p2)						\
)

/******************************************************************************/
/* NAME:        MMT_half                                                      */
/* FUNKTION:    Die rationale Zahl `1/2'.                                     */
/******************************************************************************/
#define MMT_half							\
	MCAC_half

/******************************************************************************/
/* NAME:        MMT_inc                                                       */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer aus der Typenmenge              */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "++p".                         */
/******************************************************************************/
#define MMT_inc( p )							\
(									\
	MCA_pp(p)							\
)

/******************************************************************************/
/* NAME:        MMT_init                                                      */
/* PARAMETER:   name     = Name der Funktion als String                       */
/*              options  = MMT_HOLD, MMT_NOP, MMT_REMEMBER                    */
/*              func_type= Typ der Fkt.: MMT_ANY oder [vgl. MEV_LegalOp()]    */
/* FUNKTION:    Das Makro dient als Standard-Header fuer Modulfunktionen.  Es */
/*		wird geprueft,ob der Typ des Funktionswertes in den aktuellen */
/*              Kontext des Funktionsaufrufes passt.  Die  Funktionsparameter */
/*              werden in Abhaengigkeit von flat evaluiert. Im Fehlerfall er- */
/*              folgt ein Abbruch der aktuellen Funktion.                     */
/*              Dieses Makro muss immer der erste Befehl der Funktion sein !! */
/******************************************************************************/
#define MMT_init( name, options, func_type )				\
									\
	char		*MMTV_name   = (name);				\
	S_Pointer	 MMTV_s_orig = MMTV__s;				\
	long		 MMTV_ctrl   = 0L;				\
									\
									\
	if( (options) & MMT_HOLD ) MMTV__eval_type |= MEVC_HOLD_BIT;	\
									\
	MEV_header( MMTV_s_orig, 					\
		    MEVC_NULL_DEFAULT, 					\
		    MMTV__prev_func, MMTV__eval_type, MMTV__exec,	\
		    MGSC_FLAT_ARG, 					\
		    MEVC_UNKNOWN, 					\
		    &MMTV__s, 0L, (func_type),				\
		    MEVC_NO_OVERLOAD, 0L, 0L, 0L );			\
									\
	if( !((options) & MMT_REMEMBER) ) {				\
		MEV_set_mam_control(MAMC_NO_WRITE_UP);			\
	}
	
/******************************************************************************/
/* NAME:        MMT_is_minus_one                                              */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer aus der Typenmenge              */
/*                  { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }     */
/* FUNKTION:    Das Makro liefert das Ergebnis "p == -1".                     */
/******************************************************************************/
#define MMT_is_minus_one( p )						\
(									\
	(int) MCA_eq_1(p)						\
)

/******************************************************************************/
/* NAME:        MMT_is_number                                                 */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer                                 */
/* FUNKTION:    Das Makro liefert das Ergebnis 1, wenn p aus der Typenmenge   */
/*              { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX } ist,    */
/*              und sonst 0.                                                  */
/******************************************************************************/
#define MMT_is_number( p )						\
(									\
	MMTV_l = MMT_cat(p),						\
	( MMTV_l == CAT_APM     ||					\
	  MMTV_l == CAT_COMPLEX ||					\
	  MMTV_l == CAT_FLOAT   ||					\
	  MMTV_l == CAT_INT     ||					\
	  MMTV_l == CAT_RAT     )					\
)

/******************************************************************************/
/* NAME:        MMT_is_one                                                    */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer aus der Typenmenge              */
/*                  { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }     */
/* FUNKTION:    Das Makro liefert das Ergebnis "p == 1".                      */
/******************************************************************************/
#define MMT_is_one( p )							\
(									\
	(int) MCA_eq1(p)						\
)

/******************************************************************************/
/* NAME:        MMT_is_zero                                                   */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer aus der Typenmenge              */
/*                  { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }     */
/* FUNKTION:    Das Makro liefert das Ergebnis "p == 0".                      */
/******************************************************************************/
#define MMT_is_zero( p )						\
(									\
	(int) MCA_eq0(p)						\
)

/******************************************************************************/
/* NAME:        MMT_lcm                                                       */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "kgv(p1,p2)".                  */
/******************************************************************************/
#define MMT_lcm( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_LCM)					\
)

/******************************************************************************/
/* NAME:        MMT_le                                                        */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 <= p2".                    */
/******************************************************************************/
#define MMT_le( p1, p2 )						\
(									\
	(int) MCA_le(p1,p2)						\
)

/******************************************************************************/
/* NAME:        MMT_lt                                                        */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 < p2".                     */
/******************************************************************************/
#define MMT_lt( p1, p2 )						\
(									\
	(int) MCA_lt(p1,p2)						\
)

/******************************************************************************/
/* NAME:        MMT_minus_one                                                 */
/* FUNKTION:    Die ganze Zahl `-1'.                                          */
/******************************************************************************/
#define MMT_minus_one	 						\
	MCAC_one_

/******************************************************************************/
/* NAME:        MMT_minus_two                                                 */
/* FUNKTION:    Die ganze Zahl `-2'.                                          */
/******************************************************************************/
#define MMT_minus_two	 						\
	MCAC_two_

/******************************************************************************/
/* NAME:        MMT_mod                                                       */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 % p2".                     */
/******************************************************************************/
#define MMT_mod( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_MOD)					\
)

/******************************************************************************/
/* NAME:        MMT_mult                                                      */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 * p2".                     */
/******************************************************************************/
#define MMT_mult( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_MULT)					\
)

/******************************************************************************/
/* NAME:        MMT_m2c_bool                                                  */
/* PARAMETER:   ip= (S_Pointer) Ein S_Pointer vom Typ CAT_BOOL.               */
/* FUNKTION:    Liefert 1, falls p TRUE ist und 0, sonst.                     */
/******************************************************************************/
#define MMT_m2c_bool( p )						\
(									\
	(int) MCM_equal(p,MMT_true)					\
)

/******************************************************************************/
/* NAME:        MMT_m2c_double                                                */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer vom Typ CAT_FLOAT               */
/* FUNKTION:    Liefert den C-double zum MuPAD-float (CAT_FLOAT) p.           */
/******************************************************************************/
#define MMT_m2c_double( p )						\
(									\
	MCA_cat_to_double(p)						\
)

/******************************************************************************/
/* NAME:        MMT_m2c_ident                                                 */
/* PARAMETER:   mstr = (S_Pointer) Ein S_Pointer vom Typ CAT_IDENT            */
/* FUNKTION:    Liefert eine Kopie des C-String zum Namen des MuPAD-Bezeich-  */
/*              ners mstr oder im Fehlerfall den Wert NULL.                   */
/******************************************************************************/
#define MMT_m2c_ident( p )						\
(									\
	MMT_m2c_string(p)						\
)

/******************************************************************************/
/* NAME:        MMT_m2c_ident_adr                                             */
/* PARAMETER:   mstr = (S_Pointer) Ein S_Pointer vom Typ CAT_IDENT            */
/* FUNKTION:    Liefert die Adresse des C-String zum Namen des MuPAD-Bezeich- */
/*              ners mstr. Der Zeiger wird NICHT in den Pool aufgenommen.     */
/******************************************************************************/
#define MMT_m2c_ident_adr( p )						\
(									\
	MMT_m2c_string_adr(p)						\
)

/******************************************************************************/
/* NAME:        MMT_m2c_long                                                  */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer vom Typ CAT_INT                 */
/* FUNKTION:    Liefert den C-long zum MuPAD-integer (CAT_INT) p.             */
/******************************************************************************/
#define MMT_m2c_long( p )						\
(									\
	MCA_cat_to_int(p)						\
)

/******************************************************************************/
/* NAME:        MMT_ne                                                        */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 != p2".                    */
/******************************************************************************/
#define MMT_ne( p1, p2 )						\
(									\
	MCA_neq(p1,p2)							\
)

/******************************************************************************/
/* NAME:        MMT_new_cat_expr                                              */
/* PARAMETER:   n = Anzahl der freien Ausdruckseintraege (incl. 0-tem Arg.)   */
/* FUNKTION:    Das Makro liefert einen (undef.) Ausdruck mit n Eintraegen    */
/******************************************************************************/
#define MMT_new_cat_expr( n )						\
(									\
	MTR_new_cat_expr(MMT_STORAGE,(long)(n))				\
)

/******************************************************************************/
/* NAME:        MMT_new_cat_list                                              */
/* PARAMETER:   n = Anzahl der freien Listeneintraege                         */
/* FUNKTION:    Das Makro liefert eine (undefinierte) Liste fuer n Eintraege  */
/******************************************************************************/
#define MMT_new_cat_list( n )						\
(									\
	MTR_new_cat_stat_list(MMT_STORAGE,(long)(n))			\
)

/******************************************************************************/
/* NAME:        MMT_nil                                                       */
/* FUNKTION:    Der Wert NIL.                                                 */
/******************************************************************************/
#define MMT_nil	 							\
	MUPV_CAT_NIL

/******************************************************************************/
/* NAME:        MMT_no_write_ass_mem                                          */
/* PARAMETER:   -                                                             */
/* FUNKTION:    Verhindert, dass das Funktionsergebnis in den Assoziativspei- */
/*              cher eingetragen wird und bewirkt damit, dass diese Funktion  */
/*              bei jedem Aufruf wirklich ausgefuehrt wird.                   */
/******************************************************************************/
#define MMT_no_write_ass_mem()	 					\
(									\
	MEV_set_mam_control(MAMC_NO_WRITE_UP);				\
)

/******************************************************************************/
/* NAME:        MMT_null                                                      */
/* FUNKTION:    Der Wert NULL.                                                */
/******************************************************************************/
#define MMT_null							\
	MUPV_CAT_NULL

/******************************************************************************/
/* NAME:        MMT_one                                                       */
/* FUNKTION:    Die ganze Zahl `1'.                                           */
/******************************************************************************/
#define MMT_one		 						\
	MCAC_one

/******************************************************************************/
/* NAME:        MMT_param                                                     */
/* PARAMETER:   n = (long) Index aus [0 .. MMT_count(s)]                      */
/* FUNKTION:    Liefert den n-ten Funktionsparameter.                         */
/******************************************************************************/
#define MMT_param( n )	 						\
(									\
	MMT_gop(MMTV__s,n)						\
)

/******************************************************************************/
/* NAME:        MMT_params                                                    */
/* PARAMETER:   -                                                             */
/* FUNKTION:    Liefert die Anzahl der Funktionsparameter.                    */
/******************************************************************************/
#define MMT_params()	 						\
(									\
	MMT_count(MMTV__s)-1L						\
)

/******************************************************************************/
/* NAME:        MMT_pari2cat                                                  */
/* PARAMETER:   p = (S_Pointer) Eine MuPAD-Zahl                               */
/* FUNKTION:    Wandelt das PARI-Objekt (GEN) in eine Categorie (Zahl).       */
/******************************************************************************/
#define MMT_pari2cat(p)							\
(									\
	MCA_new_cat_pari(MMT_STORAGE,p)					\
)

/******************************************************************************/
/* NAME:        MMT_power                                                     */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 ^ p2".                     */
/******************************************************************************/
#define MMT_power( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_POWER)					\
)

/******************************************************************************/
/* NAME:        MMT_puts                                                      */
/* PARAMETER:   s = (char*) C-String OHNE Formatanweisung                     */
/* FUNKTION:    Arbeitet wie  'puts' , aber gibt den String via MuPAD-        */
/*              Ausgaberoutinen aus.                                          */
/******************************************************************************/
#define MMT_puts( s )							\
	MIO_print_string(s)

/******************************************************************************/
/* NAME:        MMT_return                                                    */
/* PARAMETER:   p = Funktionswert (S_Pointer)                                 */
/* FUNKTION:    Das Makro liefert (nach einer Reinitialisierung) `p' als      */
/*              Funktionswert.                                                */
/*              BUGS: Bei strict-ANSI erfolgt manchmal die Warnung:           */
/*                    "statement not reached", wenn das Makro mit `;'         */
/*                    (leeres Kommando) abgeschlossen wurde.                  */
/******************************************************************************/
#define MMT_return( p )							\
{									\
	MMTV_t = (p);							\
	MUP_set_sign(MMTV_t,MMTV__s);					\
	MMT_exit();							\
	if( MEVV_ERRORLEVEL == MEVC_FULL_ERROR_DETECTION ) {		\
		MEV_eval( MMTV_t, MMTV__prev_func, MEVC_TYPE_CHECK );	\
	}								\
	return( MMTV_t );						\
}

/******************************************************************************/
/* NAME:        MMT_sig                                                       */
/* PARAMETER:   p = (S_Pointer) Ein S_Pointer                                 */
/* FUNKTION:    Berechnet und setzt die Signatur des S_Pointers p neu.        */
/******************************************************************************/
#define MMT_sig( p ) 							\
(									\
	MSG_sig(p)							\
)
	
/******************************************************************************/
/* NAME:        MMT_sop                                                       */
/* PARAMETER:   p1 = (S_Pointer) Ein S_Pointer mit Platz fuer Unterobjekte    */
/*              n  = (long) Index aus [0 .. MMT_count(s)]                     */
/*              p2 = (S_Pointer) Das einzuhaengende Objekt                    */
/* FUNKTION:    Besetzt den n-ten S_Pointer aus dem S_Pointer p1 mit p2.      */
/******************************************************************************/
#define MMT_sop( p1, n, p2 ) 						\
(									\
	MMTV_s = (p1),							\
	MMTV_t = (p2),							\
	MMTV_l = (long)(n),						\
	*MMMP(&MMTV_s,MMTV_l,MMMwrite) = MMTV_t,			\
	MMT_inactiv(&MMTV_s),						\
	MMTV_s								\
)

/******************************************************************************/
/* NAME:        MMT_spec                                                      */
/* PARAMETER:   p = (S_Pointer) Ein MuPAD-Datum                               */
/* FUNKTION:    Liefert den special-Eintrag des S_Pointers p.                 */
/******************************************************************************/
#define MMT_spec(p)	 						\
(									\
	MMTV_t = (p),							\
	MMTV_l = MMMZ(&MMTV_t,CATEGORY,MMMread)->spec,			\
	MMT_inactiv(&MMTV_t),						\
	MMTV_l								\
)

/******************************************************************************/
/* NAME:        MMT_sub                                                       */
/* PARAMETER:   p1, p2 = (S_Pointer) Sind S_Pointer aus der Typenmenge        */
/*                     { CAT_INT, CAT_APM, CAT_RAT, CAT_FLOAT, CAT_COMPLEX }  */
/* FUNKTION:    Das Makro liefert das Ergebnis "p1 - p2".                     */
/******************************************************************************/
#define MMT_sub( p1, p2 )						\
(									\
	MCA_calc(p1,p2,MCAC_MINUS)					\
)

/******************************************************************************/
/* NAME:        MMT_this_dom                                                  */
/* FUNKTION:    Die Funktion liest den Namen des Moduldomains aus dem zweiten */
/*              Argument von 'exec' und liefert das dazugehoerige Domain.     */
/******************************************************************************/
#define MMT_this_dom()							\
(									\
	MMT_exec_dom(MMTV__exec)					\
)

/******************************************************************************/
/* NAME:        MMT_true                                                      */
/* FUNKTION:    Der boolsche Wert `TRUE'.                                     */
/******************************************************************************/
#define MMT_true							\
	MUPV_CAT_BOOL_TRUE

/******************************************************************************/
/* NAME:        MMT_two                                                       */
/* FUNKTION:    Die ganze Zahl `2'.                                           */
/******************************************************************************/
#define MMT_two								\
	MCAC_two

/******************************************************************************/
/* NAME:        MMT_zero                                                      */
/* FUNKTION:    Die ganze Zahl `0'.                                           */
/******************************************************************************/
#define MMT_zero							\
	MCAC_zero


/*** FUNKTIONEN ***************************************************************/


/******************************************************************************/
/* NAME:        MMT_call_mf                                                   */
/* PARAMETER:   mname = Modulname, fname = Funktionsname                      */
/*              ...   = Standardargumente einer System- bzw. Modulfunktion    */
/* FUNKTION:    Ruft eine beliebige Modulfunktion auf. Dabei muss die Funk-   */
/*              tion selbst (falls noetig) die generischen Objekte laden.     */
/******************************************************************************/
static S_Pointer MMT_call_mf(	char          *mname,
				char          *fname,
				S_Pointer      s,
				long           prev_func,
				long           eval_type,
				S_Pointer      exec
			    )
{
	S_Pointer	 tmp;
	
	MMMcopy( exec );
	MMMchange( &exec );

	MMMfree( *MMMP(&exec,1L,MMMwrite) );
	*MMMP(&exec,1L,MMMwrite) = MMT_c2m_string( mname );
	MMMfree( *MMMP(&exec,2L,MMMwrite) );
	*MMMP(&exec,2L,MMMwrite) = MMT_c2m_string( fname );
	MMT_sig( exec );
	
	tmp = MDM_eval_exec( s, prev_func, eval_type, exec );
	MMT_free( exec );
	
	return( tmp );
}


/******************************************************************************/
/* NAME:        MMT_change                                                    */
/* PARAMETER:   p = (S_Pointer*) Ein Zeiger auf ein MuPAD-Objekt              */
/* FUNKTION:    Liefert eine physikalische Kopie von p.                       */
/******************************************************************************/
static void MMT_change( S_Pointer *p )
{
	MMMchange(p);
}


/******************************************************************************/
/* NAME:        MMT_copy                                                      */
/* PARAMETER:   p = (S_Pointer) Ein MuPAD-Objekt                              */
/* FUNKTION:    Liefert eine neue Instanz (Kopie) von p.                      */
/******************************************************************************/
static S_Pointer MMT_copy( S_Pointer p )
{
	MMMcopy(p);
	return(p);
}


/******************************************************************************/
/* NAME:        MMT_free                                                      */
/* PARAMETER:   p = (S_Pointer) Ein MuPAD-Objekt                              */
/* FUNKTION:    Loescht das gegebene MuPAD-Objekt bzw. einen Verweis darauf.  */
/******************************************************************************/
static S_Pointer MMT_free( S_Pointer p )
{
	MMMfree(p);
	return(p);
}


/******************************************************************************/
/* NAME:        MMT_call_eval_func                                            */
/* PARAMETER:   name = (char*) Name der Funktion                              */
/*              num  = (long)  Anzahl der Argumente                           */
/*              ...  = (S_Pointer) die Argumente                              */
/* FUNKTION:    Evaluiert die Funktion 'name' mit den gegebenen Argumenten    */
/*              und liefert das Ergebnis.                                     */
/******************************************************************************/
static S_Pointer MMT_call_eval_func( char *name, long num Mva_arglist )
{
	S_Pointer	expr, res, tmp;
	long		i, j;

	
	Mva_start( args, num );
	num++;
	expr = MTR_new_cat_expr( MMT_STORAGE, num );

	*MMMP(&expr,0L,MMMwrite) = MMT_c2m_ident( name );
	for( i = 1L; i < num; i++ ) {
		if( (tmp = Mva_arg(args,S_Pointer)) == NULL ) break;
		*MMMP(&expr,i,MMMwrite) = tmp;
		MMMcopy( tmp );
	}
	if( i < num ) {
		j = i;
		while( j < num ) {
			*MMMP(&expr,j,MMMwrite) = MMMNULL;
			j++;
		}
	}
	MMMinactiv( &expr );
	MMT_sig( expr );
	Mva_end( args );
	
	res = MEV_eval( expr, MEVC_ANY, MEVC_EVAL );
	MMMfree( expr );
	
	return( res );
}


/******************************************************************************/
/* NAME:        MMT_dom_entry                                                 */
/* PARAMETER:   domain = (S_Pointer) Eine Domain, vom Typ DOM_DOMAIN          */
/*              string = (S_Pointer) Ein String, vom Typ DOM_STRING           */
/* FUNKTION:    Das Makro liefert den Wert des Eintrages p2 im Domain p1.     */
/******************************************************************************/
static S_Pointer MMT_dom_entry( S_Pointer domain, S_Pointer string )
{
	if( MMT_cat(domain) != CAT_DOM ) {
		return( MMT_copy(MMT_fail) );
	}
	return( MEE_domain_in(domain,string) );
}


/******************************************************************************/
/* NAME:        MMT_dom_entry_c                                               */
/* PARAMETER:   domain = (S_Pointer) Eine Domain, vom Typ DOM_DOMAIN          */
/*              string = (S_Pointer) Ein String, vom Typ DOM_STRING           */
/* FUNKTION:    Das Makro liefert den Wert des Eintrages p2 im Domain p1.     */
/******************************************************************************/
static S_Pointer MMT_dom_entry_c( S_Pointer domain, char *string )
{
	S_Pointer	 tmp, result;
	
	if( MMT_cat(domain) != CAT_DOM ) {
		return( MMT_copy(MMT_fail) );
	}
	tmp = MMT_c2m_string( string );
	result = MEE_domain_in( domain, tmp );
	MMT_free( tmp );
	return( result );
}


/******************************************************************************/
/* MMT_inactiv() muss definiert werden, da MMMinactiv() ggf. ein leeres Makro */
/* ist und das darf bei der Verwendung des Komma-Operators (s.o.) nicht sein. */
/******************************************************************************/
static void MMT_inactiv( S_Pointer *p )
{ 
	p=p;
	MMMinactiv(p);
}


/******************************************************************************/
/* NAME:        MMT_is_ext_of_dom                                             */
/* PARAMETER:   domain = (CAT_DOMAIN) Das Moduldomain                         */
/*              value  = (CAT_...   ) Ein beliebiges Objekt                   */
/* FUNKTION:    Die Funktion liest den Namen des Moduldomains aus dem zweiten */
/*              Argument von 'exec'  und vergleicht ihn mit dem Domaineintrag */
/*              des Domain, auf das das erste Argument des 'ext' verweist.Bei */
/*              Gleichheit liefert die Funktion den Wert 1L, sonst 0L.        */
/******************************************************************************/
static long MMT_is_ext_of_dom ( S_Pointer domain, S_Pointer value )
{
	if( MMT_cat(value) != CAT_EXT ) {
		return( 0L );
	}
	return( MMT_equal(domain,MMT_gop(value,0L)) );
}


/******************************************************************************/
/* NAME:        MMT_m2c_string                                                */
/* PARAMETER:   mstr = (S_Pointer) Ein S_Pointer vom Typ CAT_STRING,CAT_IDENT */
/* FUNKTION:    Liefert eine Kopie des C-String zum MuPAD-String / zum Namen  */
/*              des MuPAD-Bezeichners mstr oder im Fehlerfall den Wert NULL.  */
/******************************************************************************/
static char *MMT_m2c_string( S_Pointer mstr )
{
	char		*str1, *str2;
	
	str1 = MMT_m2c_string_adr(mstr);
	MMMpoolinsert(&str);
	str2 = MMMEmalloc(1);
	if( (str2 = (char*)(MMMEmalloc(strlen(str1)+1))) == NULL ) {
		return( NULL );
	}
	strcpy(str2,str1);
	MMMpooldelete(&str1);
	return( str2 );
}


/******************************************************************************/
/* NAME:        MMT_m2c_string_adr                                            */
/* PARAMETER:   mstr = (S_Pointer) Ein S_Pointer vom Typ CAT_STRING,CAT_IDENT */
/* FUNKTION:    Liefert die Adresse des C-String zum MuPAD-String / zum Namen */
/*              des MuPAD-Bezeichners mstr. Der Zeiger wird NICHT in den Pool */
/*              aufgenommen.                                                  */
/******************************************************************************/
static char *MMT_m2c_string_adr( S_Pointer mstr )
{
	S_Pointer	 tmp;
	char		*str;
	
	tmp = *MMMP(&mstr,0L,MMMread);
	MMMinactiv(&mstr);
	str = MMMmvarr(&tmp,sizeof(STRSIZE),0L,char,MMMread);
	MMMinactiv(&tmp);
	return( str );
}


/******************************************************************************/
/* NAME:        MMT_new_cat_expr2                                             */
/* PARAMETER:   num      = (long) Anzahl der Eintraege                        */
/*              va_alist = (S_Pointer) (ggf. schon kopierte Objekte !)        */
/* FUNKTION:    Erzeugt einen S_Pointer vom Typ CAT_EXPR aus den gegebenen    */
/*              Objekten. Die Objektliste muss mindestens num Elemente ent-   */
/*              oder aber durch NULL terminiert sein.  NULL darf nicht als    */
/*              Objekt innerhalb der Liste auftreten.                         */
/******************************************************************************/
static S_Pointer MMT_new_cat_expr2( long num Mva_arglist )
{
	S_Pointer	expr, tmp;
	long		i;

	
	Mva_start( args, num );
	expr = MTR_new_cat_expr( MMT_STORAGE, num );

	for( i = 0L; i < num; i++ ) {
		if( (tmp = Mva_arg(args,S_Pointer)) == NULL ) break;
		/* MMT_sop( expr, i, tmp ); */
		*MMMP(&expr,i,MMMwrite) = tmp;
	}
	while( i < num ) {
		/* MMT_sop( expr, i, MMMNULL ); */
		*MMMP(&expr,i,MMMwrite) = MMMNULL;
		i++;
	}
	MMMinactiv( &expr );
	MMT_sig( expr );
	Mva_end( args );
	
	return( expr );
}


/******************************************************************************/
/* NAME:        MMT_new_cat_ext                                               */
/* PARAMETER:   n = Anzahl der freien DOM_EXT-Eintraege (incl. 0-tem Arg.)    */
/* FUNKTION:    Das Makro liefert ein (undef.) DOM_EXT mit n Eintraegen.      */
/******************************************************************************/
static S_Pointer MMT_new_cat_ext( long count, long size )
{
	S_Pointer	result;

	result = MMMmalloc( MMT_STORAGE, MMM_CATEGORY, 
			(long) sizeof(CATEGORY)+size,
			(long) count );
	MMMZ(&result, CATEGORY, MMMwrite)->cat  = CAT_EXT ;
	MMMZ(&result, CATEGORY, MMMwrite)->spec = NO_SYM_BIT ;
	MMMinactiv( &result ) ;
	return( result );
}


/******************************************************************************/
/* NAME:        MMT_printf                                                    */
/* PARAMETER:   format   = (char*) Formatanweisung                            */
/*              va_alist = Parametersequenz                                   */
/* FUNKTION:    Arbeitet wie  'printf' , aber gibt den String via MuPAD-      */
/*              Ausgaberoutinen aus.                                          */
/******************************************************************************/
static void MMT_printf( char *format Mva_arglist )
{	
	Mva_start( args, format );
	vsprintf( MMTV_buf, format, args );
	Mva_end  ( args );
	MIO_print_string( MMTV_buf );
}


/******************************************************************************/
/* NAME:        MMT_read_ident                                                */
/* PARAMETER:   ident = (S_Pointer) S_Pointer vom Typ CAT_IDENT, CAT_STRING   */
/* FUNKTION:    Die Funktion liefert den Wert des Bezeichners ident.  Falls   */
/*              ident nicht definiert ist, liefert die Funktion MMMNULL.      */
/******************************************************************************/
static S_Pointer MMT_read_ident( S_Pointer ident )
{
	S_Pointer	 MMTV_u;
	MSTTinfo         info;
	MEVTstamp        stamp;
	long		 procdepth;
	
	return(	MST_read_ident(
			MMT_gop(ident,0L), 
			MEVC_EVAL,MSTC_option_standard,&info,&stamp,&procdepth
		)
	);
}


/******************************************************************************/
/* NAME:        MMT_read_ident_c                                              */
/* PARAMETER:   cstr = (C-String) Name des Bezeichners, als C-String          */
/* FUNKTION:    Die Funktion liefert den Wert des Bezeichners ident.  Falls   */
/*              ident nicht definiert ist, liefert die Funktion MMMNULL.      */
/******************************************************************************/
static S_Pointer MMT_read_ident_c( char *cstr )
{
	S_Pointer	s1, s2;
	
	s1 = MMT_c2m_string(cstr);
	s2 = MMT_read_ident(s1);
	MMT_free(s1);
	return( s2 );
}


/******************************************************************************/
/* NAME:        MMT_write_ident                                               */
/* PARAMETER:   ident = (S_Pointer) S_Pointer vom Typ CAT_IDENT, CAT_STRING   */
/*              value = (S_Pointer) Der neue Wert != MMMNULL.                 */
/* FUNKTION:    Die Funktion weist dem Bezeichner ident eine Kopie des Wertes */
/*              value zu.                                                     */
/******************************************************************************/
static void MMT_write_ident( S_Pointer ident, S_Pointer value )
{
	S_Pointer	 MMTV_u;
	MSTTinfo         info;
	MEVTstamp        stamp;        
	long		 procdepth;

	MST_write_ident(
		MMT_gop(ident,0L), 
		value, MSTC_option_standard, &info, &stamp, &procdepth
	);
	MEV_refresh_ass_mem(MEVC_refreshoption_assign,info,stamp,procdepth); 
}


/******************************************************************************/
/* NAME:        MMT_write_ident_c                                             */
/* PARAMETER:   cstr = (C-String) Name des Bezeichners, als C-String          */
/*              value = (S_Pointer) Der neue Wert != MMMNULL.                 */
/* FUNKTION:    Die Funktion weist dem Bezeichner ident eine Kopie des Wertes */
/*              value zu.                                                     */
/******************************************************************************/
static void MMT_write_ident_c( char *cstr, S_Pointer value )
{
	S_Pointer	s1;
	
	s1 = MMT_c2m_string(cstr);
	MMT_write_ident(s1,value);
	MMT_free(s1);
}


/***  HILFSFUNKTIONEN UND -MAKROS *********************************************/

#endif /* __MMT_tool__ */
